@@ -4,13 +4,20 @@ from azureml.core import Workspace, Environment
 from azureml.core.conda_dependencies import CondaDependencies
 from azureml.core.compute import ComputeTarget, AmlCompute
 
-import click, os
+import click, os, time, re
 
 from azureml_ngc_tools.AzureMLComputeCluster import AzureMLComputeCluster
 from azureml_ngc_tools.cli import ngccontent
+from azureml.exceptions._azureml_exception import ProjectSystemException
 
+### SETUP LOGGING
+fileFormatter = logging.Formatter("%(asctime)s [%(levelname)-5.10s]  %(message)s")
+logger = logging.getLogger("azureml_ngc")
+logger.setLevel("DEBUG")
 
-logger = logging.getLogger(__name__)
+fileHandler = logging.FileHandler("azureml_ngc_tools.log", mode="w")
+fileHandler.setFormatter(fileFormatter)
+logger.addHandler(fileHandler)
 
 
 @click.command()
@@ -28,17 +35,30 @@ def start(login, app):
     resource_group = login_config["azureml_user"]["resource_group"]
     workspace_name = login_config["azureml_user"]["workspace_name"]
 
-    ws = Workspace(
-        workspace_name=workspace_name,
-        subscription_id=subscription_id,
-        resource_group=resource_group,
-    )
+    try:
+        ws = Workspace(
+            workspace_name=workspace_name,
+            subscription_id=subscription_id,
+            resource_group=resource_group,
+        )
+    except ProjectSystemException:
+        msg = f'\n\nThe workspace "{workspace_name}" does not exist. '
+        msg += f"Go to \n\n  "
+        msg += f"-->> https://docs.microsoft.com/en-us/azure/machine-learning/how-to-manage-workspace <<--\n\n"
+        msg += f"and create the workspace first.\n\n\n"
+        msg += f"Your current configuration: \n\n"
+        msg += f"Workspace name: {workspace_name} \n"
+        msg += f"Subscription id: {subscription_id} \n"
+        msg += f"Resource group: {resource_group}\n\n"
+
+        logger.exception(msg)
+        raise Exception(msg)
 
     verify = f"""
     Subscription ID: {subscription_id}
     Resource Group: {resource_group}
     Workspace: {workspace_name}"""
-    print(verify)
+    logger.info(verify)
 
     ### experiment name
     exp_name = login_config["aml_compute"]["exp_name"]
@@ -51,19 +71,22 @@ def start(login, app):
     ### trust but verify
     verify = f"""
     Experiment name: {exp_name}"""
-    print(verify)
+    logger.info(verify)
 
     ### GPU RUN INFO
     workspace_vm_sizes = AmlCompute.supported_vmsizes(ws)
+    pascal_volta_pattern = pattern = re.compile(
+        r"[a-z]+_nc[0-9]+[s]?_v[2,3]"
+    )  ### matches NC-series v2 and v3
     workspace_vm_sizes = [
         (e["name"].lower(), e["gpus"])
         for e in workspace_vm_sizes
-        if "_nc" in e["name"].lower()
+        if pattern.match(e["name"].lower())
     ]
     workspace_vm_sizes = dict(workspace_vm_sizes)
 
+    ### GET COMPUTE TARGET
     if vm_name in workspace_vm_sizes:
-        # vm_name = login_config["aml_compute"]["vm_name"]
         gpus_per_node = workspace_vm_sizes[vm_name]
 
         verify = f"""
@@ -72,64 +95,70 @@ def start(login, app):
     No of GPUs: {gpus_per_node}
     Priority: {vm_priority}
         """
-        print(verify)
+        logger.info(verify)
 
         ### get SSH keys
         ssh_key_pub, pri_key_file = get_ssh_keys()
 
         if ct_name not in ws.compute_targets:
-            print(f"Compute target {ct_name} does not exist...")
-            config = AmlCompute.provisioning_configuration(
-                vm_size=vm_name,
-                min_nodes=login_config["aml_compute"]["min_nodes"],
-                max_nodes=login_config["aml_compute"]["max_nodes"],
-                vm_priority=vm_priority,
-                idle_seconds_before_scaledown=login_config["aml_compute"][
-                    "idle_seconds_before_scaledown"
-                ],
-                admin_username=login_config["aml_compute"]["admin_name"],
-                admin_user_ssh_key=ssh_key_pub,
-                remote_login_port_public_access="Enabled",
+            logger.warning(f"Compute target {ct_name} does not exist...")
+            ct = createOrGetComputeTarget(
+                ws, ct_name, vm_name, vm_priority, ssh_key_pub, login_config
             )
-            ct = ComputeTarget.create(ws, ct_name, config)
-            ct.wait_for_completion(show_output=True)
         else:
-            print(
-                "    Loading pre-existing compute target {ct_name}".format(
-                    ct_name=ct_name
-                )
-            )
             ct = ws.compute_targets[ct_name]
+
+            if ct.provisioning_state == "Failed":
+                logger.warning(
+                    f"Compute target {ct_name} found but provisioning_state is showing as 'failed'..."
+                )
+                logger.warning(f"Deleting {ct_name} target and will attempt again...")
+                logger.warning(
+                    f"If this fails again check that you have enough resources in your subscription..."
+                )
+
+                ct.delete()
+                time.sleep(5)
+                ct = createOrGetComputeTarget(
+                    ws, ct_name, vm_name, vm_priority, ssh_key_pub, login_config
+                )
+            else:
+                logger.info(f"    Using pre-existing compute target {ct_name}")
     else:
-        print("Unsupported vm_size {vm_size}".format(vm_size=vm_name))
-        print("The specified vm size must be one of ...")
+        logger.exception("Unsupported vm_size {vm_size}".format(vm_size=vm_name))
+        logger.exception("The specified vm size must be one of ...")
+
         for azure_gpu_vm_size in workspace_vm_sizes.keys():
-            print("... " + azure_gpu_vm_size)
+            logger.exception("... " + azure_gpu_vm_size)
         raise Exception(
             "{vm_size} does not have Pascal or above GPU Family".format(vm_size=vm_name)
         )
 
-    environment_name = login_config["aml_compute"]["environment_name"]
-    python_interpreter = login_config["aml_compute"]["python_interpreter"]
-    conda_packages = login_config["aml_compute"]["conda_packages"]
+    env = createOrGetEnvironment(ws, login_config, app_config)
 
-    if environment_name not in ws.environments:
-        env = Environment(name=environment_name)
-        env.docker.enabled = login_config["aml_compute"]["docker_enabled"]
-        env.docker.base_image = None
-        env.docker.base_dockerfile = f'FROM {app_config["base_dockerfile"]}'
-        env.python.interpreter_path = python_interpreter
-        env.python.user_managed_dependencies = True
-        conda_dep = CondaDependencies()
+    ### UPLOAD ADDITIONAL CONTENT IF NOT EXISTS
+    for additional_content in app_config["additional_content"]["list"]:
+        url = additional_content["url"]
+        targetfile = additional_content["filename"]
+        src_path = additional_content["localdirectory"]
+        dest_path = additional_content["computedirectory"]
 
-        for conda_package in conda_packages:
-            conda_dep.add_conda_package(conda_package)
+        if app_config["additional_content"]["download_content"]:
+            ngccontent.download(url, "additional_content", targetfile)
 
-        env.python.conda_dependencies = conda_dep
-        env.register(workspace=ws)
-        evn = env
-    else:
-        env = ws.environments[environment_name]
+        if (
+            app_config["additional_content"]["unzip_content"]
+            and additional_content["zipped"]
+        ):
+            ngccontent.unzipFile(targetfile, "additional_content", src_path)
+
+        if app_config["additional_content"]["upload_content"]:
+            ngccontent.upload_data(
+                ws,
+                ws.get_default_datastore(),
+                "additional_content/" + src_path,
+                dest_path,
+            )
 
     amlcluster = AzureMLComputeCluster(
         workspace=ws,
@@ -143,8 +172,8 @@ def start(login, app):
         admin_ssh_key=pri_key_file,
     )
 
-    print(f"\n    Go to: {amlcluster.jupyter_link}")
-    print("    Press Ctrl+C to stop the cluster.")
+    logger.info(f"\n    Go to: {amlcluster.jupyter_link}")
+    logger.info("    Press Ctrl+C to stop the cluster.")
 
 
 def get_ssh_keys():
@@ -199,6 +228,60 @@ def get_ssh_keys():
     return pubkey, pri_key_file
 
 
+def createOrGetComputeTarget(
+    ws, ct_name, vm_name, vm_priority, ssh_key_pub, login_config
+):
+    config = AmlCompute.provisioning_configuration(
+        vm_size=vm_name,
+        min_nodes=login_config["aml_compute"]["min_nodes"],
+        max_nodes=login_config["aml_compute"]["max_nodes"],
+        vm_priority=vm_priority,
+        idle_seconds_before_scaledown=login_config["aml_compute"][
+            "idle_seconds_before_scaledown"
+        ],
+        admin_username=login_config["aml_compute"]["admin_name"],
+        admin_user_ssh_key=ssh_key_pub,
+        remote_login_port_public_access="Enabled",
+    )
+    ct = ComputeTarget.create(ws, ct_name, config)
+    ct.wait_for_completion(show_output=True)
+
+    if ct.provisioning_state != "Success":
+        msg = f"Failed to create the cluster..."
+        logger.exception(msg)
+        raise Exception(msg)
+    return ct
+
+
+def createOrGetEnvironment(ws, login_config, app_config):
+    environment_name = login_config["aml_compute"]["environment_name"]
+    python_interpreter = login_config["aml_compute"]["python_interpreter"]
+    conda_packages = login_config["aml_compute"]["conda_packages"]
+
+    ### CREATE OR RETRIEVE THE ENVIRONMENT
+    if environment_name not in ws.environments:
+        logger.info(f"Creating {environment_name} environment...")
+        env = Environment(name=environment_name)
+        env.docker.enabled = login_config["aml_compute"]["docker_enabled"]
+        env.docker.base_image = None
+        env.docker.base_dockerfile = f'FROM {app_config["base_dockerfile"]}'
+        env.python.interpreter_path = python_interpreter
+        env.python.user_managed_dependencies = True
+        conda_dep = CondaDependencies()
+
+        for conda_package in conda_packages:
+            conda_dep.add_conda_package(conda_package)
+
+        env.python.conda_dependencies = conda_dep
+        env.register(workspace=ws)
+        evn = env
+    else:
+        logger.info(f"    Environment {environment_name} found...")
+        env = ws.environments[environment_name]
+
+    return env
+
+
 def go():
     start()
 
